// Copyright (C) 2020  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package main

// XXX doc  (2 ways of testing: explicit + allstructs), treegen py helper

import (
	"bufio"
	"context"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"os/exec"
	"reflect"
	"regexp"
	"sort"
	"strings"
	"testing"

	"lab.nexedi.com/kirr/go123/exc"
	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"

	"github.com/stretchr/testify/require"
)

// TreeGen represents connection to running `treegen trees` server.
//
// Create it with StartTreeGen(zurl).
// - Commit(treeTopology) -> tid
type TreeGen struct {
	zurl	  string
	pysrv     *exec.Cmd      // spawned `treegen trees`
	pyin      io.WriteCloser // input to pysrv
	pyoutRaw  io.ReadCloser  // output from pysrv
	pyout     *bufio.Reader  // buffered ^^^
	treeRoot  zodb.Oid       // oid of the tree treegen works on
	head      zodb.Tid       // last made commit
}

// StartTreeGen spawns `treegen trees` server.
func StartTreeGen(zurl string) (_ *TreeGen, err error) {
	defer xerr.Contextf(&err, "treegen %s: start", zurl)

	// spawn `treegen trees`
	tg := &TreeGen{zurl: zurl}
	tg.pysrv = exec.Command("./testprog/treegen.py", "trees", zurl)
	tg.pyin, err = tg.pysrv.StdinPipe()
	if err != nil {
		return nil, err
	}
	tg.pyoutRaw, err = tg.pysrv.StdoutPipe()
	if err != nil {
		return nil, err
	}
	tg.pyout = bufio.NewReader(tg.pyoutRaw)
	tg.pysrv.Stderr = os.Stderr // no redirection

	err = tg.pysrv.Start()
	if err != nil {
		return nil, err
	}

	// wait for start message; initialize .head and .treeRoot from it
	defer func() {
		if err != nil {
			tg.Close() // ignore error
		}
	}()
	defer xerr.Context(&err, "handshake")
	start, err := tg.pyout.ReadString('\n')
	if err != nil {
		if err == io.EOF {
			err = io.ErrUnexpectedEOF
		}
		return nil, err
	}
	start = strings.TrimSuffix(start, "\n")
	defer xerr.Contextf(&err, "invalid start %q", start)

	// start @<at> tree=<root>
	startRe := regexp.MustCompile(`^start @([^ ]+) root=([^ ]+)$`)
	m := startRe.FindStringSubmatch(start)
	if m == nil {
		return nil, fmt.Errorf("unexpected format")
	}
	tg.head, err = zodb.ParseTid(m[1]) // <at>
	if err != nil {
		return nil, fmt.Errorf("tid: %s", err)
	}
	tg.treeRoot, err = zodb.ParseOid(m[2]) // <root>
	if err != nil {
		return nil, fmt.Errorf("root: %s", err)
	}

	return tg, nil
}

// Close shutdowns treegen server.
func (tg *TreeGen) Close() (err error) {
	defer xerr.Contextf(&err, "treegen %s: close", tg.zurl)

	err1 := tg.pyin.Close()
	err2 := tg.pyoutRaw.Close()
	err3 := tg.pysrv.Wait()
	return xerr.Merge(err1, err2, err3)
}

// Commit creates new commit with underlying tree changed to tree topology.
func (tg *TreeGen) Commit(tree string) (_ zodb.Tid, err error) {
	defer xerr.Contextf(&err, "treegen %s: commit %s", tg.zurl, tree)

	_, err = io.WriteString(tg.pyin, tree + "\n")
	if err != nil {
		return zodb.InvalidTid, err
	}

	reply, err := tg.pyout.ReadString('\n')
	if err != nil {
		return zodb.InvalidTid, err
	}
	reply = strings.TrimSuffix(reply, "\n")

	tid, err := zodb.ParseTid(reply)
	if err != nil {
		return zodb.InvalidTid, fmt.Errorf("invalid reply: %s", err)
	}
	tg.head = tid
	return tid, nil
}


// XGetTree loads Tree from zurl@at->obj<root>.
// Tree values must be ZBlk whose data is returned instead of references to ZBlk objects.
func XGetTree(db *zodb.DB, at zodb.Tid, root zodb.Oid) map[Key]string {
	defer exc.Contextf("%s: @%s: get tree %s", db.Storage().URL(), at, root)
	X := exc.Raiseif

	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()

	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at}); X(err)
	xztree, err := zconn.Get(ctx, root); X(err)
	ztree, ok := xztree.(*Tree)
	if !ok {
		exc.Raisef("expected %s, got %s", typeOf(ztree), typeOf(xztree))
	}

	// XXX better load through direct nodes traversal, not via bucket.next?
	// (firstbucket and .next might be broken)
	err = ztree.PActivate(ctx); X(err)
	defer ztree.PDeactivate()
	zbucket := ztree.FirstBucket()

	kv := make(map[Key]string)
	for zbucket != nil {
		err = zbucket.PActivate(ctx); X(err)
		defer zbucket.PDeactivate()

		for _, __ := range zbucket.Entryv() {
			k  := __.Key()
			xv := __.Value()
			zv, ok := xv.(zBlk)
			if !ok {
				exc.Raisef("[%d] -> %s;  want ZBlk", k, typeOf(xv))
			}

			data, _, err := zv.loadBlkData(ctx); X(err)
			kv[k] = string(data)
		}

		zbucket = zbucket.Next()
	}

	return kv
}


// xverifyΔBTail verifies how ΔBTail handles ZODB update for a tree with changes in between at1->at2.
//
// it is known that @at1 and @at2 the tree has kv1 and kv2 values correspondingly.
// it is known that for at1->at2 ZODB-level change is δZ.
func xverifyΔBTail(t *testing.T, db *zodb.DB, treeRoot zodb.Oid, at1, at2 zodb.Tid, kv1, kv2 map[Key]string, δZ *zodb.EventCommit) {
	// verify transition at1->at2 for all initial states of tracked {keys} from kv1
	keyv1 := []Key{}
	for k1 := range kv1 {
		keyv1 = append(keyv1, k1)
	}
	sort.Slice(keyv1, func(i, j int) bool {
		return keyv1[i] < keyv1[j]
	})
	for kidx := range IntSets(len(keyv1)) {
		keys := make([]Key, len(kidx))
		for _, i := range kidx {
			keys[i] = keyv1[i]
		}

		xverifyΔBTail1(t, db, treeRoot, at1,at2, kv1,kv2, δZ, keys)
	}
}

// xverifyΔBTail1 verifies how ΔBTail handles ZODB update at1->at2 from initial
// tracked state defined by initialTrackedKeys.
func xverifyΔBTail1(t *testing.T, db *zodb.DB, treeRoot zodb.Oid, at1,at2 zodb.Tid, kv1,kv2 map[Key]string, δZ *zodb.EventCommit, initialTrackedKeys []Key) {
	X := exc.Raiseif
	assert := require.New(t)
	fmt.Println("verify1", initialTrackedKeys)

	txn, ctx := transaction.New(context.Background())
	defer txn.Abort()

	// zconn, δbtail @at1 with initial tracked set
	zconn, err := db.Open(ctx, &zodb.ConnOptions{At: at1});	X(err)
	δbtail := NewΔBtail(zconn.At())

	xtree, err := zconn.Get(ctx, treeRoot);	X(err)
	ztree := xtree.(*Tree)

	for _, k := range initialTrackedKeys {
		path := []Node{}
		_, _, err = ztree.VGet(ctx, k, func(node Node) {
			path = append(path, node)
		}); X(err)

		δbtail.Track(path)
	}

	// δbtail <- δZ
	δB := δbtail.Update(δZ)
	assert.Equal(δB.Rev, δZ.Tid)

	d12 := kvdiff(kv1, kv2)
	fmt.Println("d12:", d12)
	if len(d12) == 0 {
		assert.Equal(len(δB.ByRoot), 0)
		return
	}

	assert.Equal(len(δB.ByRoot), 1)

	// XXX assert δB.keys == treeRoot
	// δT = δB[treeRoot]

	// all keys from d12/tracked are present in δT
	//for (k,v) in d12:
	//	if k in initialTrackedKeys:
	//		assert k in δT
	//		assert δT[k] == v
	//	else:
	//		if k in δT:
	//			assert δT[k] == v
	//
	// XXX all keys from δT are present in d12
}

func TestΔBTail(t *testing.T) {
	X := exc.Raiseif

	// test known cases going through tree1 -> tree2 -> ...
	testv := []string {
		"T/B:",
		"T/B1:a",
		"T/B2:b",
		// XXX
	}
	// direct  tree_i -> tree_{i+1} -> _{i+2} ...   plus
	// reverse ... tree_i <- _{i+1} <- _{i+2}
	for i := len(testv)-2; i >= 0; i-- {
		testv = append(testv, testv[i])
	}

	work, err := ioutil.TempDir("", "δBTail"); X(err)
	defer func() {
		err := os.RemoveAll(work); X(err)
	}()
	tg, err := StartTreeGen(work + "/1.fs"); X(err)
	defer func() {
		err := tg.Close(); X(err)
	}()

	zstor, err := zodb.Open(context.Background(), tg.zurl, &zodb.OpenOptions{
					ReadOnly: true,
		      }); X(err)
	defer func() {
		err := zstor.Close(); X(err)
	}()

	db := zodb.NewDB(zstor)
	defer func() {
		err := db.Close(); X(err)
	}()

	// XCommitTree calls tg.Commit and returns δZ corresponding to committed transaction.
	XCommitTree := func(tree string) *zodb.EventCommit {
		defer exc.Contextf("commit %s", tree)

		watchq := make(chan zodb.Event)
		at0 := zstor.AddWatch(watchq)
		defer zstor.DelWatch(watchq)

		tid, err := tg.Commit(tree); X(err)

		if !(tid > at0) {
			exc.Raisef("treegen -> %s  ; want > %s", tid, at0)
		}

		zevent := <-watchq
		δZ := zevent.(*zodb.EventCommit)
		if δZ.Tid != tid {
			exc.Raisef("treegen -> %s  ; watchq -> %s", tid, δZ)
		}
		return δZ
	}

	/*
	XXX := func(tree1, tree2 string) {
		// XXX commit tree1
		// XXX load tree1 -> get keys for kv1
		//
		// XXX for all set of {keys} from kv1:
		//	// prepare initial state with {keys} tracked
		//	zconn  = db.open(at)
		//	δbtail = NewΔBTail(zconn.At())
		//	for k in keys:
		//		v, path = tree.VGet(k)
		//		δbtail.Track(path)
		//
		//	commit(->tree2)
		//	δZ = ...
		//
		//	δB = δbtail.Update(δZ)
		//
		//	verify δB has all keys as expected for tracked set

		_, err := tg.Commit(tree1); X(err)
		kv1 := XGetTree(db, tg.head, tg.treeRoot)
		fmt.Println(kv1)
		_, err = tg.Commit(tree2); X(err)
		kv2 := XGetTree(db, tg.head, tg.treeRoot)
		fmt.Println(kv2)
	}
	*/

	at1 := tg.head
	kv1 := XGetTree(db, at1, tg.treeRoot)
	for _, tree := range testv {
		δZ := XCommitTree(tree) // tree2
		at2 := δZ.Tid
		kv2 := XGetTree(db, at2, tg.treeRoot)

		xverifyΔBTail(t, db, tg.treeRoot, at1,at2, kv1,kv2, δZ)

		at1 = at2
		kv1 = kv2
	}

	// TODO reverse tree_i <- tree_{i+1}
}

func TestΔBTreeAllStructs(t *testing.T) {
	// XXX given (kv1, kv2) test on automatically generated (tree1 -> tree2)
}



// IntSets generates all sets of integers in range [0,N)
func IntSets(N int) chan []int {
	ch := make(chan []int)
	go intSets(ch, 0, N)
	return ch
}

// intSets generates all sets of integers in range [lo,hi)
func intSets(ch chan []int, lo, hi int) {
	ch <- nil // ø

	if lo < hi {
		for i := lo; i < hi; i++ {
			chTail := make(chan []int)
			go intSets(chTail, i+1, hi)
			for tail := range chTail {
				ch <- append([]int{i}, tail...) // i + tail
			}
		}
	}

	close(ch)
}

func TestIntSets(t *testing.T) {
	got := [][]int{}
	for is := range IntSets(3) {
		got = append(got, is)
	}

	I := func(v ...int) []int { return v }
	want := [][]int{I(),
			I(0), I(0,1), I(0,1,2), I(0,2),
			I(1), I(1,2),
			I(2),
	}

	if !reflect.DeepEqual(got, want) {
		t.Fatalf("error:\ngot:  %v\nwant: %v", got, want)
	}
}


// kvdiff returns differens in between kv1 and kv2.
var DEL = "ø"
type kvdelta1 struct {
	k Key
	v string // DEL means delete
}
func kvdiff(kv1, kv2 map[Key]string) []kvdelta1 {
	delta := []kvdelta1{}
	keys := SetKey{}
	for k := range kv1 { keys.Add(k) }
	for k := range kv2 { keys.Add(k) }
	keyv := []Key{}
	for k := range keys { keyv = append(keyv, k) }
	sort.Slice(keyv, func(i,j int) bool {
		return keyv[i] < keyv[j]
	})

	for _, k := range keyv {
		v1, ok := kv1[k]
		if !ok { v1 = DEL }
		v2, ok := kv2[k]
		if !ok { v2 = DEL }
		if v1 != v2 {
			delta = append(delta, kvdelta1{k,v2})
		}
	}

	return delta
}
